use std::{
    collections::HashMap,
    sync::{
        Arc,
        mpsc::{Receiver, RecvTimeoutError, Sender},
    },
    thread::JoinHandle,
    time::Duration,
};

/// Status receiver of a dead man switch
pub struct DeadManReceiver {
    rx: Receiver<bool>,
}

impl DeadManReceiver {
    /// Checks if a dead man switch is alive
    pub fn alive(&self) -> bool {
        let res = self.rx.recv_timeout(Duration::from_millis(400));
        if let Err(err) = res {
            return match err {
                RecvTimeoutError::Timeout => true,
                RecvTimeoutError::Disconnected => false,
            };
        }

        false
    }
}

/// A Dead Man Switch
pub struct DeadManSwitch {
    pub lifeline: Sender<bool>,
}

impl DeadManSwitch {
    /// Create a new Dead Man Switch
    ///
    /// The Dead Man Switch will notify the receiver and mark itself dead after it got dropped (for example in a panic)
    pub fn new() -> (Self, DeadManReceiver) {
        let (tx, rx) = std::sync::mpsc::channel();

        (Self { lifeline: tx }, DeadManReceiver { rx })
    }
}

impl Drop for DeadManSwitch {
    fn drop(&mut self) {
        log::info!("I am DEAD man 💀");
        let _ = self.lifeline.send(true);
    }
}

/// Service Manager
///
/// This will manage and keep background services running-
pub struct ServiceManager<T: Fn(DeadManSwitch) -> () + 'static> {
    pub services: HashMap<String, Arc<T>>,
}

impl<T: Fn(DeadManSwitch) -> () + 'static + Send + Sync> ServiceManager<T> {
    pub fn new() -> Self {
        Self {
            services: HashMap::new(),
        }
    }

    /// Register a new background service
    pub fn register(&mut self, service_name: &str, service: T) {
        self.services
            .insert(service_name.to_string(), Arc::new(service));
    }

    fn run(self) -> ! {
        let mut threads = HashMap::new();

        // Initial start
        for (service, f) in self.services.iter() {
            log::info!("Starting service {service}");
            let (dms, dmr) = DeadManSwitch::new();
            let f = Arc::clone(&f);
            threads.insert(
                service.clone(),
                (
                    std::thread::spawn(move || {
                        f(dms);
                    }),
                    dmr,
                ),
            );
        }

        // Monitor loop
        loop {
            let mut threads_new = HashMap::new();

            for (service, (t, dm)) in &mut threads {
                let alive = dm.alive();
                log::debug!(
                    "Checking up on {service} -- {}",
                    if alive { "ALIVE" } else { "DEAD" }
                );

                if !alive {
                    log::warn!("Service {service} died. Restarting...");

                    while !t.is_finished() {
                        log::debug!("waiting for thread");
                    }

                    let f = Arc::clone(self.services.get(service).unwrap());

                    let (dms, dmr) = DeadManSwitch::new();
                    threads_new.insert(
                        service.clone(),
                        (
                            std::thread::spawn(move || {
                                f(dms);
                            }),
                            dmr,
                        ),
                    );
                }
            }

            threads.extend(threads_new);
        }
    }

    /// Start the services
    ///
    /// This returns a ThreadHandle for the service monitor thread.
    pub fn spawn(self) -> JoinHandle<()> {
        std::thread::spawn(|| self.run())
    }
}
